import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Tuple
from uuid import UUID

from langchain_core.documents import Document as LangchainDocument

from langcommunity.callback.displays.llamaindex.models.record import (
    LLamaIndexRetrieverRecord,
    LLamaIndexRetrieverState,
)

logger = logging.getLogger(__name__)

try:
    from llama_index.core.callbacks.base_handler import BaseCallbackHandler  # type: ignore
    from llama_index.core.callbacks.schema import CBEventType, EventPayload  # type: ignore
    from llama_index.core.schema import NodeWithScore  # type: ignore
except ImportError:
    raise ImportError()


class BaseAsyncLLamaIndexCallbackHandler(BaseCallbackHandler, ABC):
    """
    Base callback handler that can be used to track event starts and ends.

    This class is a bridge between the LlamaIndex callback system and the Langchain callback system.
    """

    verbose: bool = True
    record: Optional[LLamaIndexRetrieverRecord] = None
    event_starts_to_ignore: Tuple = tuple()
    event_ends_to_ignore: Tuple = tuple()

    # Abstract
    # ---
    @abstractmethod
    def llama_index_on_retriever(self, retriever: LLamaIndexRetrieverRecord, **kwargs: Any) -> None:
        """
        Abstract method to handle a retriever event.

        Args:
            retriever: The retriever record.
            **kwargs: Additional keyword arguments.

        Returns:
            None
        """
        pass

    # Bridge callbacks
    # ---
    async def llama_index_on_prepare(
        self,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: List[str] | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Callback method called when preparing for LLamaIndex retriever.
        """
        self.record = LLamaIndexRetrieverRecord(
            run_id=run_id,
            parent_run_id=parent_run_id,
            state=LLamaIndexRetrieverState.PREPARED,
            tags=tags,
        )
        self.llama_index_on_retriever(retriever=self.record, **kwargs)

    # Callbacks
    # ---
    def on_event_start(
        self,
        event_type: CBEventType,
        payload: Optional[Dict[str, Any]] = None,
        event_id: str = "",
        parent_id: str = "",
        **kwargs: Any,
    ) -> None:
        """
        Callback method called when an event starts.
        """
        if event_type != CBEventType.QUERY:
            return

        if not self.record:
            raise ValueError(
                "Retriever record is not initialized. Please Add `on_retriever_prepare_llama_index` to retrieve with managers callback langchaing to initialize the retriever record."  # noqa: E501
            )

        self.record.update(
            query=payload[EventPayload.QUERY_STR.value],
            state=LLamaIndexRetrieverState.START,
        )
        self.llama_index_on_retriever(retriever=self.record, **kwargs)

    def on_event_end(
        self,
        event_type: CBEventType,
        payload: Optional[Dict[str, Any]] = None,
        event_id: str = "",
        **kwargs: Any,
    ) -> None:
        """
        Callback method called when an event ends.
        """
        if not self.record:
            raise ValueError(
                "Retriever record is not initialized. Please Add `on_retriever_prepare_llama_index` to "
                "retrieve with managers callback langchaing to initialize the retriever record."
            )

        # NOTE: Take the last one from Retrieve else Rerank
        if event_type == CBEventType.RETRIEVE or event_type == CBEventType.RERANKING:
            # TODO: can be exception
            if payload and EventPayload.NODES in payload:
                documents_nodes: List[NodeWithScore] = payload[EventPayload.NODES]

                documents = []
                for node in documents_nodes:
                    metadata = {**node.metadata}
                    metadata["score"] = node.score
                    documents.append(
                        LangchainDocument(
                            page_content=node.text,
                            metadata=node.metadata,
                        )
                    )
                self.record.update(
                    state=LLamaIndexRetrieverState.END,
                    documents=documents,
                )
            if payload and EventPayload.EXCEPTION in payload:
                logger.error(
                    payload,
                    extra={
                        "title": "[ERROR] Exception in LLAMMA" + " : " + self.__class__.__name__,
                        "verbose": True,
                    },
                )

        elif event_type == CBEventType.QUERY:
            self.llama_index_on_retriever(retriever=self.record, **kwargs)

    def start_trace(self, trace_id: Optional[str] = None) -> None:
        """
        Callback method called when an overall trace is launched.
        """
        pass

    def end_trace(
        self,
        trace_id: Optional[str] = None,
        trace_map: Optional[Dict[str, List[str]]] = None,
    ) -> None:
        """
        Callback method called when an overall trace is exited.
        """
        pass



import json
import logging
from typing import Dict, List, Optional

from langchain_core.chat_history import BaseChatMessageHistory
from langchain_core.messages import BaseMessage, message_to_dict, messages_from_dict

logger = logging.getLogger(__name__)

try:
    from pymongo import MongoClient, errors  # type: ignore[unused-ignore] # noqa: F401
except ImportError:
    raise ImportError("langchain_groq library is not installed. Please install `poetry add pymongo motor`.")


DEFAULT_DBNAME = "chat_history"
DEFAULT_COLLECTION_NAME = "message_store"
DEFAULT_SESSION_ID_KEY = "chat_session_id"
DEFAULT_HISTORY_KEY = "history"


class MongoDBChatMessageHistory(BaseChatMessageHistory):
    """Chat message history that stores history in MongoDB.

    Args:
        connection_string: connection string to connect to MongoDB
        session_id: arbitrary key that is used to store the messages
            of a single chat session.
        database_name: name of the database to use
        collection_name: name of the collection to use
        session_id_key: name of the field that stores the session id
        history_key: name of the field that stores the chat history
        create_index: whether to create an index on the session id field
        index_kwargs: additional keyword arguments to pass to the index creation
    """

    def __init__(
        self,
        session_id: str,
        collection_name: str,
        client: MongoClient,
        database_name: str = DEFAULT_DBNAME,
        *,
        session_id_key: str = DEFAULT_SESSION_ID_KEY,
        history_key: str = DEFAULT_HISTORY_KEY,
        create_index: bool = True,
        index_kwargs: Optional[Dict] = None,
    ):
        self.session_id = session_id
        self.database_name = database_name
        self.collection_name = collection_name
        self.session_id_key = session_id_key
        self.history_key = history_key
        self.client = client

        self.db = self.client[database_name]
        self.collection = self.db[collection_name]

        if create_index:
            index_kwargs = index_kwargs or {}
            self.collection.create_index(self.session_id_key, **index_kwargs)

    @property
    def messages(self) -> List[BaseMessage]:  # type: ignore
        """Retrieve the messages from MongoDB"""
        try:
            cursor = self.collection.find({self.session_id_key: self.session_id})
        except errors.OperationFailure as error:
            logger.error(error)

        if cursor:
            items = [json.loads(document[self.history_key]) for document in cursor]
        else:
            items = []

        messages = messages_from_dict(items)
        return messages

    def add_message(self, message: BaseMessage) -> None:
        """Append the message to the record in MongoDB"""
        try:
            self.collection.insert_one(
                {
                    self.session_id_key: self.session_id,
                    self.history_key: json.dumps(message_to_dict(message)),
                }
            )
        except errors.WriteError as err:
            logger.error(err)

    def clear(self) -> None:
        """Clear session memory from MongoDB"""
        try:
            self.collection.delete_many({self.session_id_key: self.session_id})
        except errors.WriteError as err:
            logger.error(err)



from __future__ import annotations

import asyncio
from typing import Any, AsyncIterator, Union

from langfoundation.callback.base.records.token import (
    TokenStream,
    TokenStreamState,
)

from langcommunity.callback.displays.fastapi.models.complete import (
    CompleteStreamResponse,
)
from langcommunity.callback.displays.fastapi.models.error import ErrorStreamResponse
from langcommunity.callback.displays.fastapi.models.token import TokenStreamResponse
from langcommunity.callback.displays.mixin.llamaindex.callback import (
    BaseAsyncDisplayMixinCallbackHandler,
)


class AsyncFastAPICallbackHandler(BaseAsyncDisplayMixinCallbackHandler):
    """Callback handler that returns an async iterator.

    This class implements the callbacks for the FastAPI integration.
    """

    queue: asyncio.Queue[Union[TokenStreamResponse, CompleteStreamResponse, ErrorStreamResponse]]

    def __init__(
        self,
        should_cumulate_token: bool = False,
    ) -> None:
        """
        Args:
            should_cumulate_token (bool): Whether to cumulate the token. Defaults to False.
        """
        self.should_cumulate_token = should_cumulate_token
        self.queue = asyncio.Queue()

    # Callbacks
    # ----
    async def on_token_stream(
        self,
        token: TokenStream,
        **kwargs: Any,
    ) -> None:
        match token.state:
            case TokenStreamState.ERROR:
                self.queue.put_nowait(
                    ErrorStreamResponse(
                        id=token.id,
                        error=str(token.error) if token.error else "error",
                    )
                )
            case _:
                if token.state == TokenStreamState.END:
                    self.queue.put_nowait(
                        CompleteStreamResponse(
                            id=token.id,
                        )
                    )
                else:
                    self.queue.put_nowait(
                        TokenStreamResponse(
                            id=token.id,
                            token=token.token,
                        )
                    )

    # Token Iterator
    # ----
    async def atoken_iterator(self) -> AsyncIterator[str]:
        """Returns an async iterator over the token stream."""
        while True:
            reponse = await self.queue.get()
            if isinstance(reponse, CompleteStreamResponse):
                yield "data: " + "[DONE]"
                self.queue.task_done()
                break
            elif isinstance(reponse, ErrorStreamResponse):
                yield "data: " + reponse.json()
                yield "data: " + "[DONE]"
                self.queue.task_done()
                break
            else:
                yield "data: " + reponse.json()



from typing import List, Tuple, Union

import torch
from gary.services.similarity.base import BaseSimilarity
from sentence_transformers import SentenceTransformer
from sentence_transformers.util import cos_sim, semantic_search


class Similarity(BaseSimilarity):
    """
    # Similarity

    This class provides functionality for computing semantic similarity between a query sentence and a list of documents.
    It utilizes pre-trained models from the Hugging Face Model Hub, specifically designed for sentence similarity tasks.

    The default model used is 'sentence-transformers/sentence-t5-base', which has demonstrated good performance in various tasks.

    Other available models can be explored on the Hugging Face website under the 'sentence-similarity' pipeline tag:
    https://huggingface.co/models?pipeline_tag=sentence-similarity&sort=downloads

    Some notable models include:
    - 'sentence-transformers/all-MiniLM-L12-v2'
    - 'sentence-transformers/paraphrase-MiniLM-L6-v2'
    - sentence-transformers/all-mpnet-base-v2 (400mb)
    - sentence-transformers/sentence-t5-base (120mb)
    - BAAI/bge-large-zh-v1.5 (https://huggingface.co/BAAI/bge-large-en-v1.5) (1.32Gb) (Good result)
    - mixedbread-ai/mxbai-embed-large-v1
    - etc.

    Usage:
    ```
    similarity = Similarity(sentence_transformers_name="sentence-transformers/sentence-t5-base")
    query = "Ma maman est belle"
    documents = ["Ma maman est partie", "Ma maman est magnifique"]
    similarity_results = similarity.semantic_search(query, documents)
    similarity_scores = similarity.calculate_similarity(query, documents)
    ```
    """

    def __init__(self, sentence_transformers_name: str = "mixedbread-ai/mxbai-embed-large-v1") -> None:
        """
        Initialize the Similarity class.

        Parameters:
        - sentence_transformers_name (str): The name or path of the pre-trained sentence transformers model.
                                                Default is "sentence-transformers/sentence-t5-base".
        """
        self._device: torch.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self._sentence_transformers_name: str = sentence_transformers_name
        self._embeddings_model: Union[SentenceTransformer, None] = None

    # Public Methods
    # ---

    async def configure(self) -> None:
        """
        Initialize the sentence embeddings model.
        """
        if self._embeddings_model is None:
            self._embeddings_model = await self.load_model()

        self._embeddings_model.to(self._device)
        self._embeddings_model.eval()

    async def load_model(self) -> SentenceTransformer:
        return SentenceTransformer(self._sentence_transformers_name, cache_folder="./cache")

    def calculate_similarity(self, query: str, documents: List[str]) -> List[float]:
        """
        Calculate the cosine similarity between the query and a list of documents.

        Parameters:
        - query (str): The query sentence.
        - documents (List[str]): List of document sentences.

        Returns:
        - List[float]: List of similarity scores between the query and each document.
        """
        # Encode sentences
        query_embedding, documents_embeddings = self._encode_query_documents(query, documents)

        # Calculate cosine similarity
        return [cos_sim(query_embedding, emb).item() for emb in documents_embeddings]

    def semantic_search(self, query: str, documents: List[str], top_k: int = 5) -> List[Tuple[int, float]]:
        """
        Perform semantic search between the query and a list of documents.

        Parameters:
        - query (str): The query sentence.
        - documents (List[str]): List of document sentences.
        - top_k (int): The number of top results to return. Default is 5.

        Returns:
        - List[Tuple[int, float]]: A list of tuples containing the index of the document
        and its similarity score with the query.
        """
        # Encode the query and documents
        query_embedding, documents_embeddings = self._encode_query_documents(query, documents)

        # Perform semantic search
        search_results = semantic_search(query_embedding, documents_embeddings, top_k=top_k)

        # Extract indices and scores from search results
        results = [(result["corpus_id"], result["score"]) for result in search_results[0]]
        return results

    # Private Methods
    # ---

    def _encode(self, sentence: str) -> torch.Tensor:
        """
        Encode a sentence into its embedding representation.

        Parameters:
        - sentence: The input sentence.

        Returns:
        - torch.Tensor: The embedding representation of the input sentence.
        """
        if self._embeddings_model is None:
            raise ValueError("Embeddings model not initialized. Please call setup() first.")
        with torch.no_grad():
            sentence_embedding = self._embeddings_model.encode(sentence, convert_to_tensor=True)
        return sentence_embedding.to(self._device)

    def _encode_query_documents(self, query: str, documents: List[str]) -> Tuple[torch.Tensor, List[torch.Tensor]]:
        """
        Encode the query and a list of documents into their respective embeddings.

        Parameters:
        - query (str): The query sentence.
        - documents (List[str]): List of document sentences.

        Returns:
        - Tuple[torch.Tensor, List[torch.Tensor]]: A tuple containing the embedding of the query
                                                    and a list of embeddings for the documents.
        """
        # Encode the query and documents
        query_embedding = self._encode(query)
        documents_embeddings = [self._encode(doc) for doc in documents]
        return query_embedding, documents_embeddings



from __future__ import annotations

from typing import Any, List

from langfoundation.callback.base.records.agent import AgentRecord
from langfoundation.callback.base.records.retriever import (
    RetrieverRecord,
)
from langfoundation.callback.base.records.token import (
    TokenStream,
    TokenStreamState,
)
from langfoundation.callback.base.records.tool import ToolRecord

from langcommunity.callback.displays.mixin.llamaindex.callback import (
    BaseAsyncDisplayMixinCallbackHandler,
)


class AsyncDisplayStoreCallbackHandler(BaseAsyncDisplayMixinCallbackHandler):
    """
    A callback handler to store AI messages in a list.
    """

    display_ai_messages: List[str] = []

    def __init__(
        self,
        verbose: bool = False,
    ):
        """
        Initialize the callback handler.

        Args:
            verbose: Whether to print out logs.
        """
        self.display_ai_messages = []
        self.verbose = verbose
        self.should_cumulate_token = True

    def clear(self) -> None:
        """Clear the stored AI messages."""
        self.display_ai_messages = []

    # Stream
    # ----
    async def on_token_stream(
        self,
        token: TokenStream,
        tags: List[str] | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Handle a token stream event.

        If the token stream is not ended, return immediately.
        """
        if token.state != TokenStreamState.END:
            return

        if tags:
            display_parser = self._find_display_parser_by_tags(tags)
            token_to_append = display_parser.parse(token.cumulate_tokens) if display_parser else token.cumulate_tokens
        else:
            token_to_append = token.cumulate_tokens

        self.display_ai_messages.append(token_to_append)

    # Tool
    # ----
    async def on_tool(self, tool: ToolRecord, **kwargs: Any) -> None:
        """Handle a tool event."""
        pass

    # Agent
    # ----
    async def on_agent(self, agent: AgentRecord, **kwargs: Any) -> None:
        """Handle an agent event."""
        pass

    # Retriever
    # ----
    async def on_retriever(self, retriever: RetrieverRecord, **kwargs: Any) -> None:
        """Handle a retriever event."""
        pass

    def non_async_on_retriever(self, retriever: RetrieverRecord, **kwargs: Any) -> None:
        """Handle a retriever event in a non-async manner."""
        pass

    # Feedback
    # ----
    async def on_feedback(self, feedback: str, **kwargs: Any) -> None:
        """Handle a feedback event."""
        pass



from typing import Any, Dict

from pydantic import BaseModel


class FrozenConfig(BaseModel):
    value: Dict[str, Any]

    def __hash__(self) -> int:
        def _hash_dict(d: Dict) -> frozenset:
            return frozenset((k, _hash_dict(v) if isinstance(v, dict) else v) for k, v in d.items())

        return hash(_hash_dict(self.value))



from __future__ import annotations

import json
import logging
from typing import Any, Dict

from pydantic import BaseModel

logger = logging.getLogger(__name__)


class Formatter:
    @classmethod
    def format(cls, value: Any, indent: int = 0) -> str:
        return Formatter().format_any(value, indent)

    @classmethod
    def dump(cls, value: Any, indent: int = 0) -> None:
        logger.debug(Formatter().format_any(value, indent))

    def __init__(self) -> None:
        self.types: Dict = {}
        self.indent_char = "  "
        self.line_ending_char = "\n"
        self.set_formatter(BaseModel, self._format_base_model)
        self.set_formatter(str, self._format_json_string)
        self.set_formatter(dict, self._format_dict)
        self.set_formatter(list, self._format_list)
        self.set_formatter(tuple, self._format_tuple)
        self.set_formatter(object, self._format_object)

    def set_formatter(self, obj: Any, callback: Any) -> None:
        self.types[obj] = callback

    def format_any(self, value: Any, indent=0) -> str:
        for type_formatter, formatter in self.types.items():
            if isinstance(value, type_formatter):
                return formatter(value, indent)
        return self._format_object(value, indent)

    def _format_object(self, value: Any, indent: int) -> str:
        return repr(value)

    def _format_dict(self, value: Any, indent: int) -> str:
        items = [
            f"{self.line_ending_char}{self.indent_char * (indent + 1)}"
            f"{repr(key)}: {self.types.get(type(value[key]), self.format_any)(value[key], indent + 1)}"
            for key in value
        ]
        return "{" + ",".join(items) + f"{self.line_ending_char}{self.indent_char * indent}" + "}"

    def _format_list(self, value: Any, indent: int) -> str:
        items = [
            f"{self.line_ending_char}{self.indent_char * (indent + 1)}" f"{self.types.get(type(item), self.format_any)(item, indent + 1)}"
            for item in value
        ]
        return "[" + ",".join(items) + f"{self.line_ending_char}{self.indent_char * indent}" + "]"

    def _format_tuple(self, value: Any, indent: int) -> str:
        items = [
            f"{self.line_ending_char}{self.indent_char * (indent + 1)}" f"{self.types.get(type(item), self.format_any)(item, indent + 1)}"
            for item in value
        ]
        return "(" + ",".join(items) + f"{self.line_ending_char}{self.indent_char * indent}" + ")"

    def _format_json_string(self, value: str, indent: int) -> str:
        try:
            json_dict = json.loads(value)
            return self._format_dict(json_dict, indent)
        except Exception:
            return self._format_object(value, indent)

    def _format_base_model(self, value: BaseModel, indent: int) -> str:
        _dict = self._format_dict(dict(value), indent)
        _dict = value.__class__.__name__ + _dict.replace('"', "").replace(":", "=").replace("{", "(").replace("}", ")")
        return _dict



from abc import ABC, abstractmethod
from typing import List, Tuple


class BaseSimilarity(ABC):
    @abstractmethod
    def calculate_similarity(self, query: str, documents: List[str]) -> List[float]:
        """
        Calculate the cosine similarity between the query and a list of documents.

        Parameters:
        - query (str): The query sentence.
        - documents (List[str]): List of document sentences.

        Returns:
        - List[float]: List of similarity scores between the query and each document.
        """

    @abstractmethod
    def semantic_search(self, query: str, documents: List[str], top_k: int = 5) -> List[Tuple[int, float]]:
        """
        Perform semantic search between the query and a list of documents.

        Parameters:
        - query (str): The query sentence.
        - documents (List[str]): List of document sentences.
        - top_k (int): The number of top results to return. Default is 5.

        Returns:
        - List[Tuple[int, float]]: A list of tuples containing the index of the document
        and its similarity score with the query.
        """



from __future__ import annotations

from typing import Any, Callable, Optional

from langfoundation.callback.base.records.agent import (
    AgentRecord,
    AgentState,
)
from langfoundation.callback.base.records.retriever import (
    RetrieverRecord,
    RetrieverState,
)
from langfoundation.callback.base.records.token import (
    TokenStream,
    TokenStreamState,
)
from langfoundation.callback.base.records.tool import (
    ToolRecord,
    ToolsState,
)

from langcommunity.callback.displays.mixin.llamaindex.callback import BaseAsyncDisplayMixinCallbackHandler
from langcommunity.utils.debug.formatter import Formatter

try:
    import streamlit as st  # type: ignore
    from streamlit.delta_generator import DeltaGenerator  # type: ignore
except ImportError:
    raise ImportError()


class AsyncStreamlitCallbackHandler(BaseAsyncDisplayMixinCallbackHandler):
    def __init__(
        self,
        msg_container: DeltaGenerator,
        chain_thought_container: Optional[DeltaGenerator] = None,
        on_start_stream_callback: Optional[Callable[..., None]] = None,
        max_tool_input_str_length: int = 60,
        max_document_str_length: int = 400,
        separator="- - -",
        verbose: bool = True,
    ):
        self._msg_container = msg_container
        self._chain_thought_container = chain_thought_container
        self._on_output_start = on_start_stream_callback

        self._max_tool_input_str_length = max_tool_input_str_length
        self._max_document_str_length = max_document_str_length
        self._separator = separator

        self.verbose = verbose
        self.should_cumulate_token = True

        self._llm_token_stream_placeholder = None
        self._status = None
        self._display_token = ""

    # Stream
    # ----
    async def on_token_stream(
        self,
        token: TokenStream,
        **kwargs: Any,
    ) -> None:
        with self._msg_container:
            if self._on_output_start:
                self._on_output_start()
                self._on_output_start = None

            if self._llm_token_stream_placeholder is None:
                self._llm_token_stream_placeholder = st.empty()

            self._display_token += token.token
            spacer_indicator = "" if token.state == TokenStreamState.END else "▌"
            display_text = self._display_token + spacer_indicator
            self._llm_token_stream_placeholder.markdown(display_text)

    # Tool
    # ----
    async def on_tool(self, tool: ToolRecord, **kwargs: Any) -> None:
        if not self._chain_thought_container:
            return

        with self._status.container():
            if tool.state == ToolsState.REQUEST:
                self._status.update(label=f"**Tool Request:** {tool.name}", state="running")
                st.markdown(f"**Tool Request:**{Formatter.format(tool.input_str)}\n{self._separator}")

            if tool.state == ToolsState.START:
                self._status.update(label=f"**Tool Start:** {tool.name}", state="running")
                st.markdown(f"**Tool Input:** {Formatter.format(tool.input_str)}\n{self._separator}")

            elif tool.state == ToolsState.END:
                self._status.update(label=f"**Tool End:** {tool.name}", state="running")
                st.markdown(f"**Tool Ouput:** {Formatter.format(tool.output)}\n{self._separator}")

            elif tool.state == ToolsState.ERROR:
                self._status.update(label=f"**Tool Error:** {tool.name}", expanded=True, state="error")
                st.exception(tool.error)
                st.markdown(f"**Tool Error:** {Formatter.format(tool.error)}\n{self._separator}")

    async def on_agent(self, agent: AgentRecord, **kwargs: Any) -> None:
        if not self._chain_thought_container:
            return

        if agent.state == AgentState.THOUGHT:
            with self._chain_thought_container:
                self._status = st.status(label="**Agent Start**", state="running")
                with self._status.container():
                    lines = agent.log.strip().split("\n")
                    result = []
                    for line in lines:
                        try:
                            key, value = line.strip().split(":")
                            result.append(f"**Agent {key}:** {value}")
                        except Exception:
                            result.append(f"{line}")
                    result_text = "\n\n".join(result)
                    st.markdown(f"{result_text}\n{self._separator}")

        elif agent.state == AgentState.END:
            # NOTE: Special Agent, status can be null when Angent answer directly without tools
            status = self._status if self._status else st.status(label="**Agent Start**", state="running")
            with status.container():
                status.update(label="**Agent**", state="complete")
                st.markdown(f"**Agent Result:** {Formatter.format(agent.result)}")
            self._status = None

    async def on_retriever(self, retriever: RetrieverRecord, **kwargs: Any) -> None:
        self._dislay_on_retriever(retriever)

    def non_async_on_retriever(self, retriever: RetrieverRecord, **kwargs: Any) -> None:
        self._dislay_on_retriever(retriever)

    async def on_feedback(self, feedback: str, **kwargs: Any) -> None:
        with self._chain_thought_container:
            with st.status(label="**Feedback**", state="complete").container():
                st.markdown(feedback)

    def _dislay_on_retriever(
        self,
        retriever: RetrieverRecord,
    ) -> None:
        if retriever.state == RetrieverState.START:
            with self._chain_thought_container:
                self._status = st.status(label="**Retrieve**", state="running")

        elif retriever.state == RetrieverState.END:
            doc_samples = "\n".join(
                [
                    f"**Chunk {index}:** "
                    + document.page_content[: self._max_document_str_length]
                    + "..."
                    + "\n\n"
                    + "**Meta Data:**\n"
                    + f"```\n\n{Formatter.format(document.metadata)}\n\n```"
                    + "\n\n"
                    + "----"
                    for index, document in enumerate(retriever.documents)
                ]
            )
            with self._status.container():
                self._status.update(state="complete")
                st.markdown(doc_samples)

            self._status = None


def _convert_newlines(text: str) -> str:
    """Convert newline characters to markdown newline sequences
    (space, space, newline).
    """
    return text.replace("\n", "  \n")



from __future__ import annotations

import asyncio
import logging
import typing
from typing import AsyncIterable, Coroutine

from langcommunity.callback.displays.fastapi.callback import AsyncFastAPICallbackHandler

logger = logging.getLogger(__name__)

try:
    from fastapi.responses import StreamingResponse as FastApiStreamingResponse  # type: ignore
    from starlette.background import BackgroundTask  # type: ignore
except ImportError:
    raise ImportError()


class CallbackStreamResponse(FastApiStreamingResponse):
    """
    A special response that run an invoke and send a callback response to the client.

    This class is a wrapper around FastAPI's StreamingResponse. It will run the invoke
    and send the callback response to the client as an event stream.
    """

    def __init__(
        self,
        invoke: Coroutine,
        callback: AsyncFastAPICallbackHandler,
        status_code: int = 200,
        headers: typing.Mapping[str, str] | None = None,
        media_type: str | None = "text/event-stream",
        background: BackgroundTask | None = None,
    ) -> None:
        """
        Initialize the response.

        Args:
        invoke: The invoke to run.
        callback: The callback to send the response to.
        status_code: The status code of the response.
        headers: The headers of the response.
        media_type: The media type of the response.
        background: The background task to run.
        """
        super().__init__(
            content=CallbackStreamResponse.send_message(callback, invoke),
            status_code=status_code,
            headers=headers,
            media_type=media_type,
            background=background,
        )

    @staticmethod
    async def send_message(callback: AsyncFastAPICallbackHandler, invoke: Coroutine) -> AsyncIterable[str]:
        """
        Send the message to the callback.

        Args:
        callback: The callback to send the message to.
        invoke: The invoke to run.

        Yields:
        The message to send to the client.
        """
        asyncio.create_task(invoke)

        async for token in callback.atoken_iterator():
            yield token



from __future__ import annotations

import asyncio
import datetime
import os
import time
from enum import auto
from typing import Awaitable, Callable, Dict, List, Optional, Union

from deepeval.dataset import EvaluationDataset
from deepeval.evaluate import TestResult
from deepeval.metrics import (
    AnswerRelevancyMetric,
    BaseMetric,
    ContextualRecallMetric,
    FaithfulnessMetric,
)
from deepeval.models.base_model import DeepEvalBaseLLM
from deepeval.test_case import LLMTestCase
from deepeval.utils import get_or_create_event_loop
from gary.debug.formatter import Formatter
from pydantic import BaseModel
from strenum import UppercaseStrEnum


class RetrievalEvaluationType(UppercaseStrEnum):
    TRIAD = auto()
    FAITHFULNESS_ANSWER_RELEVANCE = auto()


class RAGEvaluator(BaseModel):
    name: str
    evaluation_type: RetrievalEvaluationType
    metrics: List[BaseMetric]
    test_results: List[TestResult] = []

    class Config:
        arbitrary_types_allowed = True

    # Public methods
    # ---

    def clean(self):
        self.test_results = []

    def run_evals(
        self,
        invoker: Union[
            Callable[[Dict], LLMTestCase],
            Callable[[Dict], LLMTestCase],
            Callable[[Dict], Awaitable[LLMTestCase]],
            Callable[[Dict], Awaitable[LLMTestCase]],
        ],
        eval_questions: List[Dict[str, str]],
        cleanup_previous_results: bool = True,
        run_async: bool = True,
        batch_size: int = 1,
        batch_interval: float = 0.2,
        store_dir: Optional[str] = None,
    ) -> List[TestResult]:
        if cleanup_previous_results:
            self.clean()

        self.test_results = [] if self.test_results is None else self.test_results
        for metric in self.metrics:
            metric.async_mode = run_async

        file_path: Optional[str] = None
        if store_dir:
            file_path = self._create_test_results_file(store_dir)

        dataset = EvaluationDataset()
        for use_case in eval_questions:
            test_case: LLMTestCase
            if asyncio.iscoroutinefunction(invoker):
                loop = get_or_create_event_loop()
                test_case = loop.run_until_complete(invoker(use_case))
            else:
                test_case = invoker(use_case)  # type: ignore

            dataset.add_test_case(test_case)

            if len(dataset.test_cases) > batch_size and file_path:
                self._evaluate(dataset, file_path)
                dataset = EvaluationDataset()
                time.sleep(batch_interval)

        # Case where not enter depending of the batch, still data
        if dataset.test_cases and file_path:
            self._evaluate(dataset, file_path)

        if file_path:
            self._write_metric_pass_rates(self.test_results, file_path)
            self._write_summary(self.test_results, file_path)

        return self.test_results

    def _evaluate(
        self,
        dataset: EvaluationDataset,
        file_path: str,
    ):
        evaluation_results = dataset.evaluate(self.metrics)
        for test_result in evaluation_results:
            self._write_test_result(test_result, file_path)

        self.test_results += evaluation_results

    def _create_test_results_file(self, store_dir: str) -> str:
        ext = ".txt"
        file_path = store_dir + "/" + self.name + ext

        # Set the initial suffix to 1
        suffix = 1

        # Loop until we find a file name that does not exist
        while os.path.isfile(file_path):
            # Increment the suffix
            suffix += 1

            # Generate a new file name with the suffix
            file_name = self.name + "_{}".format(suffix) + ext

            # Generate a new file path with the new file name
            file_path = store_dir + "/" + file_name

        # Get the current date and time
        current_datetime = datetime.datetime.now()

        # Format the date and time as a string
        current_datetime_str = current_datetime.strftime("%Y-%m-%d %H:%M:%S")

        # Create the file and open it for writing
        with open(file_path, "a") as file_object:
            file_object.write(current_datetime_str)
            file_object.write("\n")
        return file_path

    def _write_summary(self, test_results: List[TestResult], file_path: str) -> None:
        # Initialize counters
        total_success = 0
        total_error = 0
        total_results = len(test_results)
        # Prepare data for the table
        for test in test_results:
            total_success += test.success

            for metric in test.metrics:
                if metric.error:
                    total_error += 1
                break

        # Calculate means
        if total_results != 0:
            mean_success = total_success / total_results
        else:
            mean_success = 0

        headers = ["Metric", "Value"]
        data = [
            ["Total tests", total_results],
            ["Mean success", "{:.2%}".format(mean_success)],
            ["Total error", total_error],
        ]
        with open(file_path, "a") as file_object:
            file_object.write("\n")
            file_object.write("Summary")
            row_format = "{:<20} {:<15}"
            if headers is not None:
                file_object.write(row_format.format(*headers) + "\n")
                file_object.write("-" * 35 + "\n")

            for row in data:
                file_object.write(row_format.format(*row) + "\n")
            file_object.write("\n" + "=" * 70 + "\n")

    def _write_test_result(self, test_result: TestResult, file_path: str) -> None:
        with open(file_path, "a") as file_object:
            file_object.write("\n")
            file_object.write("=" * 70 + "\n")
            file_object.write("\n")
            file_object.write("Metrics Summary\n")
            file_object.write("\n")
            for metric in test_result.metrics:
                successful = True
                if metric.error is not None:
                    successful = False
                else:
                    # This try block is for user defined custom metrics,
                    # which might not handle the score == undefined case elegantly
                    try:
                        if not metric.is_successful():
                            successful = False
                    except Exception:
                        successful = False

                if not successful:
                    file_object.write(
                        (
                            f"❌ {metric.__name__}:\n"
                            f"  - score: {metric.score}\n"
                            f"  - threshold: {metric.threshold}\n"
                            f"  - strict: {metric.strict_mode}\n"
                            f"  - evaluation model: {metric.evaluation_model}\n"
                            f"  - reason: {metric.reason}\n"
                            f"  - error: {metric.error}\n\n"
                        )
                    )
                else:
                    file_object.write(
                        (
                            f"✅ {metric.__name__}:\n"
                            f"  - score: {metric.score}\n"
                            f"  - threshold: {metric.threshold}\n"
                            f"  - strict: {metric.strict_mode}\n"
                            f"  - evaluation model: {metric.evaluation_model}\n"
                            f"  - reason: {metric.reason}\n"
                            f"  - error: {metric.error}\n\n"
                        )
                    )
                if metric.score_breakdown:
                    for metric_name, score in metric.score_breakdown.items():
                        file_object.write(f"      - {metric_name} (score: {score})\n")

            formatter = Formatter()
            file_object.write("\n")
            file_object.write("For test case:\n")
            file_object.write(f"  - input: {test_result.input}\n")
            file_object.write(f"  - actual output: {test_result.actual_output}\n")
            file_object.write(f"  - expected output: {test_result.expected_output}\n")
            file_object.write(f"  - context: {formatter.format(test_result.context)}\n")
            file_object.write(f"  - retrieval context: {formatter.format(test_result.retrieval_context)}\n")

    def _write_metric_pass_rates(self, test_results: List[TestResult], file_path: str) -> None:
        metric_counts = {}
        metric_successes = {}

        for result in test_results:
            for metric in result.metrics:
                metric_name = metric.__class__.__name__
                if metric_name not in metric_counts:
                    metric_counts[metric_name] = 0
                    metric_successes[metric_name] = 0
                metric_counts[metric_name] += 1
                if metric.success:
                    metric_successes[metric_name] += 1

        metric_pass_rates = {metric: (metric_successes[metric] / metric_counts[metric]) for metric in metric_counts}

        with open(file_path, "a") as file_object:
            file_object.write("\n" + "=" * 70 + "\n")
            file_object.write("\n")
            file_object.write("Overall Metric Pass Rates\n")
            for metric, pass_rate in metric_pass_rates.items():
                file_object.write(f"{metric}: {pass_rate:.2%} pass rate\n")
            file_object.write("\n" + "=" * 70 + "\n")

    # Instance
    # ---

    @classmethod
    def from_rag_triad_evaluation(
        cls,
        name: str,
        eval_llm: Union[DeepEvalBaseLLM, str],
        context_relevance_threshold=0.7,
        faithfulness_threshold=0.7,
        answer_relevance_threshold=0.7,
    ) -> RAGEvaluator:
        """
        Build RetrievalEvalBuilder for RAG triad Evaluation

        The RAG triad is made up of 3 evaluations:
        1. Contextual Recall
        2. Faithfulness
        3. Answer Relevance

        1. Contextual Recall:
        Is the retrieved context capture the context general or general information necessary to generate the expected output?

        A higher contextual recall score indicates that the retrieval system is better at capturing all general relevant information from input.
        It evaluates the retrieved context aligns with the expected output.
        Any missing or irrelevant information in the context could lead to hallucinations or incorrect responses.

        2. Faithfulness:
        Is the response supported by the context?

        Evaluating whether the Response factually aligns with the contents of the Retrieval

        After the context is retrieved, it is then formed into an answer by an LLM.
        LLMs are often prone to straying from the facts provided, exaggerating or expanding to a correct-sounding answer.

        3. Answer Relevance:
        Is the response relevant to the query?

        Evaluating how relevant the Response of the LLM is compared to the provided Input

        Lastly, the response still needs to helpfully answer the original question.
        Evaluating the relevance of the final response to the user input.

        Args:
        - name (str): Name of the test
        - eval_llm (str | DeepEvalBaseLLM): Provider for language model, str for GPT or custom model like mistral
        - app_id (str): Application ID.
        - context_relevance_threshold (float): threshold for test success
        - faithfulness_threshold (float): threshold for test success
        - answer_relevance_threshold (float): threshold for test success

        Returns:
        - RetrievalEvalBuilder: RetrievalEvalBuilder instance configured for RAG triad evaluation.
        """  # noqa: E501

        contextual_metric = ContextualRecallMetric(threshold=context_relevance_threshold, model=eval_llm, include_reason=True)
        faithfulness_metric = FaithfulnessMetric(threshold=faithfulness_threshold, model=eval_llm, include_reason=True)
        answer_relevancy_metric = AnswerRelevancyMetric(threshold=answer_relevance_threshold, model=eval_llm, include_reason=True)

        metrics = [contextual_metric, faithfulness_metric, answer_relevancy_metric]

        return cls(name=name, evaluation_type=RetrievalEvaluationType.TRIAD, metrics=metrics)

    @classmethod
    def from_faithfulness_and_answer_relevance_evaluation(
        cls,
        name: str,
        eval_llm: Union[DeepEvalBaseLLM, str],
        faithfulness_threshold=0.7,
        answer_relevance_threshold=0.7,
    ) -> RAGEvaluator:
        """
        Build RetrievalEvalBuilder for RAG triad Evaluation

        The RAG triad is made up of 3 evaluations:
        1. Contextual Recall
        2. Faithfulness
        3. Answer Relevance

        2. Faithfulness:
        Is the response supported by the context?

        Evaluating whether the Response factually aligns with the contents of the Retrieval

        After the context is retrieved, it is then formed into an answer by an LLM.
        LLMs are often prone to straying from the facts provided, exaggerating or expanding to a correct-sounding answer.

        3. Answer Relevance:
        Is the response relevant to the query?

        Evaluating how relevant the Response of the LLM is compared to the provided Input

        Lastly, the response still needs to helpfully answer the original question.
        Evaluating the relevance of the final response to the user input.

        Args:
        - name (str): Name of the test
        - eval_llm (str | DeepEvalBaseLLM): Provider for language model, str for GPT or custom model like mistral
        - app_id (str): Application ID.
        - context_relevance_threshold (float): threshold for test success
        - faithfulness_threshold (float): threshold for test success
        - answer_relevance_threshold (float): threshold for test success

        Returns:
        - RetrievalEvalBuilder: RetrievalEvalBuilder instance configured for RAG triad evaluation.
        """

        faithfulness_metric = FaithfulnessMetric(threshold=faithfulness_threshold, model=eval_llm, include_reason=True)
        answer_relevancy_metric = AnswerRelevancyMetric(threshold=answer_relevance_threshold, model=eval_llm, include_reason=True)

        metrics = [faithfulness_metric, answer_relevancy_metric]

        return cls(
            name=name,
            evaluation_type=RetrievalEvaluationType.FAITHFULNESS_ANSWER_RELEVANCE,
            metrics=metrics,
        )



from __future__ import annotations

from typing import Any, Dict, List, Optional, cast

from langchain.memory.chat_memory import BaseChatMemory
from langchain.memory.summary import SummarizerMixin
from langchain_core.chat_history import (
    BaseChatMessageHistory,
    InMemoryChatMessageHistory,
)
from langchain_core.messages import BaseMessage, get_buffer_string
from pydantic import Field


class ConversationSummaryChatMemory(BaseChatMemory, SummarizerMixin):
    """Buffer with summarizer for storing conversation memory."""

    trigger_summary_token_limit: int = Field(
        description="Maximum number of tokens in the buffer before triggering summarization.",
        default=1000,
    )
    history_max_summary_buffer_token_limit: int = Field(
        description="Maximum number of tokens take from history for summarization buffer.",
        default=2000,
    )
    summary_last_messages_include_limit: int = Field(
        description="Number of most recent messages to include in the summary after pruning.",
        default=2,
    )

    memory_key: str = Field(
        description="Key used for accessing the memory buffer.",
        default="history",
    )

    chat_summurized: BaseChatMessageHistory = Field(
        description="History of chat messages that have been summarized.",
        default_factory=InMemoryChatMessageHistory,
    )

    @property
    def memory_variables(self) -> List[str]:
        """Will always return list of memory variables.

        :meta private:
        """
        return [self.memory_key]

    @property
    def summury_or_history_messages(self) -> List[BaseMessage]:
        """String buffer of memory."""
        if self.chat_summurized.messages:
            return self.chat_summurized.messages
        else:
            return self.chat_memory.messages

    @property
    def summury_message(self) -> Optional[BaseMessage]:
        """String buffer of memory."""
        if self.chat_summurized.messages:
            return self.chat_summurized.messages[0]
        else:
            return None

    def load_memory_variables(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Return history buffer."""
        buffer = self.summury_or_history_messages
        if self.return_messages:
            final_buffer: Any = buffer
        else:
            final_buffer = get_buffer_string(buffer, human_prefix=self.human_prefix, ai_prefix=self.ai_prefix)
        return {self.memory_key: final_buffer}

    async def aload_memory_variables(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Asynchronously return key-value pairs given the text input to the chain."""
        raise NotImplementedError()

    def summarize(self) -> None:
        """Summarize the chat history"""
        buffer = self.chat_memory.messages
        curr_buffer_length = self.llm.get_num_tokens_from_messages(buffer)
        if curr_buffer_length > self.trigger_summary_token_limit:
            pruned_memory = []
            while curr_buffer_length > self.history_max_summary_buffer_token_limit:
                pruned_memory.append(buffer.pop(0))
                curr_buffer_length = self.llm.get_num_tokens_from_messages(buffer)
            else:
                pruned_memory = buffer

            moving_summary_buffer: str = cast(str, self.summury_message.content if self.summury_message else "")
            moving_summary_buffer = self.predict_new_summary(pruned_memory, moving_summary_buffer)
            first_messages: List[BaseMessage] = [self.summary_message_cls(content=moving_summary_buffer)]

            buffer = first_messages + buffer[-self.summary_last_messages_include_limit :]
            self.clear_summarized()
            self.add_summarized_messages(buffer)

    def add_summarized_messages(
        self,
        messages: List[BaseMessage],
    ) -> None:
        """Append a list of messages to summarize"""
        self.chat_summurized.add_messages(messages)

    def clear_summarized(self) -> None:
        """Clear a list of summarized messages"""
        self.chat_summurized.clear()

    def clear(self) -> None:
        self.clear_summarized()
        self.chat_memory.clear()



from __future__ import annotations

from abc import ABC
from typing import Any

from langfoundation.callback.base.base import (
    BaseAsyncDisplayCallbackHandler,
)
from langfoundation.callback.base.records.agent import AgentRecord
from langfoundation.callback.base.records.retriever import (
    RetrieverRecord,
)
from langfoundation.callback.base.records.token import (
    TokenStreamState,
)
from langfoundation.callback.base.records.tool import ToolRecord

from langcommunity.callback.displays.llamaindex.callback import (
    BaseAsyncLLamaIndexCallbackHandler,
)
from langcommunity.callback.displays.llamaindex.models.record import (
    LLamaIndexRetrieverRecord,
    LLamaIndexRetrieverState,
    LLamaIndexStreamState,
)


class BaseAsyncDisplayMixinCallbackHandler(BaseAsyncDisplayCallbackHandler, BaseAsyncLLamaIndexCallbackHandler, ABC):
    """
    A callback handler for asynchronous base display with mixed functionality.

    This class extends from Non Langchain to Langchain CallbackHandler.

    It provides a bridge between non-Langchain (LLama Index) and Langchain.
    """

    # Bridge LLama Index to Langchain
    # ----

    def llama_index_on_retriever(self, retriever: LLamaIndexRetrieverRecord, **kwargs: Any) -> None:
        """
        Handles the retrieval of LLama index.

        Bridge LLama Index to Langchain
        """
        if retriever.state == LLamaIndexRetrieverState.START:
            self.non_async_on_retriever_start(
                query=retriever.query,
                run_id=retriever.run_id,
                parent_run_id=retriever.parent_run_id,
                tags=retriever.tags,
                **kwargs,
            )
        elif retriever.state == LLamaIndexRetrieverState.END:
            self.non_async_on_retriever_end(
                documents=retriever.documents,
                tags=retriever.tags,
                **kwargs,
            )

    async def on_tool(self, tool: ToolRecord, **kwargs: Any) -> None:
        pass

    async def on_agent(self, agent: AgentRecord, **kwargs: Any) -> None:
        pass

    async def on_retriever(self, retriever: RetrieverRecord, **kwargs: Any) -> None:
        pass

    def non_async_on_retriever(self, retriever: RetrieverRecord, **kwargs: Any) -> None:
        pass

    async def on_feedback(self, feedback: str, **kwargs: Any) -> None:
        pass

    def _to_stream_token(self, state: LLamaIndexStreamState) -> TokenStreamState:
        if state == LLamaIndexStreamState.STREAM:
            return TokenStreamState.STREAM
        elif state == LLamaIndexStreamState.START:
            return TokenStreamState.START
        elif state == LLamaIndexStreamState.END:
            return TokenStreamState.END



from datetime import datetime
from functools import cache
from typing import Optional

from langchain_community.tools.ddg_search import DuckDuckGoSearchRun
from langchain_community.tools.wikipedia.tool import WikipediaQueryRun
from langchain_community.utilities.wikipedia import WikipediaAPIWrapper
from langchain_core.tools import BaseTool, StructuredTool, Tool
from pydantic import BaseModel, Field

from langcommunity.tools.python_ast_repl_tool import PythonAstREPLTool
from langcommunity.tools.you import YahooFinanceStockPriceTool

###################################
# Tools
###################################
# TODO: For tool inject use a special << >>  to indicate format the value


def get_yahoo_finance_news_tool() -> StructuredTool:
    """Look up things on wikipedia."""

    class YahooInput(BaseModel):
        """Input for the yahoo_finance_news tool."""

        query: str = Field(
            description="Input query should be only one ticker symbol separate or enter multiple tickers separated by a comma. Example: 'AAPL, MSFT, NVDA' or 'AAPL'."  # noqa: E501
        )

    @cache
    def run(
        query: str,
    ) -> str:
        return YahooFinanceStockPriceTool().run(query)

    return StructuredTool.from_function(
        func=run,
        name="YahooFinanceStockPrice",
        description="Useful for obtaining stock price information, this tool accepts ticker symbol such as 'AAPL' for Apple, 'MSF' for Microsoft, or 'NVDA' for NVIDIA.",  # noqa: E501
        args_schema=YahooInput,
    )


def get_duckduckgo_search_tool() -> StructuredTool:
    class DuckDuckGoSearchInput(BaseModel):
        """Look up things online."""

        query: str = Field(description="Input should be a search query.")

    def handle_error(error) -> str:
        return "The following errors occurred during tool execution:" + error.args[0] + "Please try another tool."

    @cache
    def run(
        query: str,
    ) -> str:
        return DuckDuckGoSearchRun(handle_tool_error=handle_error).run(query)

    return StructuredTool.from_function(
        func=run,
        name="DuckDuckGoSearch",
        description="Useful for when you need to answer questions about current events",
        args_schema=DuckDuckGoSearchInput,
    )


def get_wiki_tool(verbose: bool = False) -> StructuredTool:
    """Look up things on wikipedia."""

    class WikiInput(BaseModel):
        """Input for the yahoo_finance_news tool."""

        query: str = Field(description="Input should be a search query. Examples: 'Macron' or 'History of France'")

    @cache
    def run(
        query: str,
    ) -> str:
        return WikipediaQueryRun(
            description=(
                "Useful for when you need to answer general questions about "
                "people, places, companies, facts, historical events, or other subjects."
            ),
            api_wrapper=WikipediaAPIWrapper(top_k_results=2, doc_content_chars_max=2048),  # type: ignore
            args_schema=WikiInput,
        ).run(query)

    return StructuredTool.from_function(
        func=run,
        name="Wikipedia",
        description=(
            "Useful for when you need to answer general questions about "
            "people, places, companies, facts, historical events, or other subjects."
        ),
        args_schema=WikiInput,
    )


def get_date_time_tool() -> BaseTool:
    def get_date(nothing) -> str:
        return datetime.now().strftime("%Y-%m-%d %H:%M")

    return Tool(
        name="Date-time",
        func=get_date,
        description=("Useful to ask the current date and time." "Not Input needed."),
    )


###################################
# Code
###################################


###################################
# Python REPL
###################################


class PythonInputs(BaseModel):
    query: str = Field(prompt="Code snippet to run")


def get_python_tool(description: Optional[str] = None) -> StructuredTool:
    """Look up things on wikipedia."""

    @cache
    def run(
        query: str,
    ) -> str:
        class PythonREPL:
            def __init__(self):
                self.local_vars = {}
                self.python_tool = PythonAstREPLTool()

            def run(self, code: str) -> str:
                output = str(self.python_tool.run(code))
                if output == "":
                    return "Executed successfully."
                else:
                    return output

        python_repl = PythonREPL()
        return python_repl.run(query)

    default_prompt = (
        "A Python shell.\n"
        "This is a Python shell interface designed for executing Python commands.\n"
        "Ensure that your input adheres to proper Python syntax.\n"
        "Make sure it does not look abbreviated before using it in your answer.\n"
        "Python executed will return:\n"
        "- If successful, the result of the executed code or msg success if result of the executed code is empty.\n"
        "- Otherwise, it will display the Python error encountered."
    )
    return StructuredTool.from_function(
        func=run,
        name="PythonShell",
        description=description or default_prompt,
        args_schema=PythonInputs,
    )



from __future__ import annotations

import sys
from typing import Any

from langfoundation.callback.base.records.token import (
    TokenStream,
    TokenStreamState,
)

from langcommunity.callback.displays.mixin.llamaindex.callback import (
    BaseAsyncDisplayMixinCallbackHandler,
)

try:
    from halo import Halo  # type: ignore
except ImportError:
    raise ImportError()


class AsyncCLICallbackHandler(BaseAsyncDisplayMixinCallbackHandler):
    """
    A callback handler for the CLI display that uses the halo library for a spinner.

    This class provides methods to handle various events related to display actions,
    such as LLM (Language Model) events, tool events, retriever events, and agent events.
    """

    spinner: Halo

    def __init__(
        self,
        should_cumulate_token: bool = False,
    ) -> None:
        """
        Initialize the callback handler.

        Args:
            should_cumulate_token: Whether to cumulate tokens.
        """
        self.spinner = Halo(spinner="dots")
        self.spinner.start()
        self.should_cumulate_token = should_cumulate_token

    async def on_token_stream(
        self,
        token: TokenStream,
        **kwargs: Any,
    ) -> None:
        """
        Handle a stream token event.

        Args:
            token: The token to handle.
        """
        if token.state == TokenStreamState.START:
            self.spinner.stop()

        sys.stdout.write(token.token)

        if token.state == TokenStreamState.END:
            sys.stdout.write("\n\n·················································\n\n")



from enum import auto
from typing import List, Optional
from uuid import UUID

from langchain_core.documents import Document as LangchainDocument
from langfoundation.callback.base.records.base import BaseRecord
from strenum import LowercaseStrEnum


class LLamaIndexRetrieverState(LowercaseStrEnum):
    """
    Enum class representing the possible states of a LLamaIndex retriever record.
    """

    PREPARED = auto()
    START = auto()
    END = auto()


class LLamaIndexStreamState(LowercaseStrEnum):
    """
    Enum class representing the possible states of a stream record for a LLamaIndex retriever.
    """

    START = auto()
    STREAM = auto()
    END = auto()


class LLamaIndexRetrieverRecord(BaseRecord):
    """
    Represents a retriever record.

    Attributes:
        run_id (UUID): The ID of the run.
        parent_run_id (Optional[UUID], optional): The ID of the parent run. Defaults to None.
        query (str): The query string.
        state (RetrieverState): The state of the retriever.
        documents (Optional[List[Document]], optional): The list of documents. Defaults to None.
    """

    run_id: Optional[UUID] = None
    parent_run_id: Optional[UUID] = None
    tags: List[str] | None = None

    query: Optional[str] = None
    state: LLamaIndexRetrieverState

    documents: Optional[List[LangchainDocument]] = None
